#include <stdio.h>
#include <sys/types.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <ctype.h>
#include <string.h>
#include <time.h>

#include "blocks.h"

#ifndef WIN32
#include <unistd.h>
#include <utime.h>
#include <dirent.h>
#else
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <sys/utime.h>
#include <direct.h>
#include <io.h>
#define utime(X,T) _utime(X,T)
#define utimbuf _utimbuf
#define unlink(F) _unlink(F)
#define rmdir(D) _rmdir(D)
#define access(P,M) _access(P,M)
#define R_OK 02
#define W_OK 04
#define mkdir(P,M) _mkdir(P)
#endif

//----------------------------------------------------------------------------
// block cache code...
//----------------------------------------------------------------------------

// 16384 blocks == 1 Gb of cache
static unsigned char *all_blocks=0;
static long block_num=0;
static unsigned long max_blocks=16384;

#define MAX_INTERESTING_BLOCKS 128
static time_t interesting_block_time[MAX_INTERESTING_BLOCKS];
static time_t interesting_block_check_time[MAX_INTERESTING_BLOCKS];
static unsigned char interesting_blocks[MAX_INTERESTING_BLOCKS][2][16];
static long interesting_block_num=0;

// erase the real timestamp on a file or directory
// TODO: find utime() replacement that works on directories under Windoze
static void erase_timestamp(char *filename)
{
    struct utimbuf buf;
	memset(&buf,0x00,sizeof(buf));
    buf.actime=961000987;
    buf.modtime=961000987;
    utime(filename,&buf);
}

// write a block into the block cache...
long write_block(BLOCK *block,unsigned long *key,unsigned long *id,int do_write)
{
    char block_type=block->block_type;
    if (block_type!='H' && block_type!='D')
        return -999; // invalid block
    
    unsigned long h[4],last_block[4],err=0,q;
    
    // store last data block ID from header blocks for
    // use later in out 'interesting blocks' list.
    if (do_write && block_type=='H')
    {
        long filesize=(block->size[0]<<24) +
            (block->size[1]<<16) +
            (block->size[2]<<8)  +
            (block->size[3]);
        long n=(filesize+sizeof(block->payload.data)-1)/
            sizeof(block->payload.data);
		if (n>sizeof(block->payload.block_ids)/16)
			n=sizeof(block->payload.block_ids)/16;
        read_key(block->payload.block_ids[n-1],last_block);
    }
    
    // clear any routing info...
    memset(block->send_route,0x00,sizeof(block->send_route));
    memset(block->return_route,0x00,sizeof(block->return_route));
  
    // scramble the block before writing it to disk
    if (key)
        scramble_buffer(block->payload.data,sizeof(block->payload.data),key);
    hash(block->payload.data,sizeof(block->payload.data),h);
   
//if (do_write) printf("write_block type=%c ID=%s\n",
//	   block->block_type,
//	   tohex(h));

    copy_key(passkey,tmpkey);
    scramble_buffer((unsigned char *)block,
		            ((char *)&block->payload.data)-((char *)block),
					tmpkey);
    
    // filename is based on hash of payload in hex
    char filename[256],*hx=tohex(h);
    sprintf(filename,"%s/%.2s/%s",cache_dir,hx,hx);
    if (id) copy_key(h,id);
    
    if (!do_write)
        return 0; // physical write not requested.
    
	// check if block already exists in cache...
	unsigned char idbytes[16];
	write_key(idbytes,h);
	for (q=0;q<max_blocks;q++)
	{
		if (all_blocks[16*q]==idbytes[0])
		{
			if (memcmp(&all_blocks[16*q+1],&idbytes[1],15)==0)
				return 0; // block exists... TODO: add to head of list ??
		}
	}
    FILE *fd;
//    if (fd=fopen(filename,"rb");)
//    {
//        fclose(fd);
//        return 0; // block exists... TODO: add to head of list ??
//    }

    // write the the block to a file
    if (!(fd=fopen(filename,"wb")))
        return -1; // open failed!
    if (fwrite((void *)block,sizeof(*block),1,fd)!=1)
        err=1;
    fclose(fd);
    
    if (err)
        return -2; // write failed.
    
    // try to change the timestamp on the file to hide when
    // it was really created.
    erase_timestamp(filename);
    
    // update our list of blocks
    long z;
    for (z=0;z<16;z++)
    {
        if (all_blocks[block_num*16+z])
        {
            // we have to destroy a block so we can insert
            unsigned long id[4];
            read_key(&all_blocks[block_num*16],id);
            char name[256],*hx=tohex(id);
            sprintf(name,"%s/%.2s/%s",cache_dir,hx,hx);
            unlink(name);
            break;
        }
    }
    write_key(&all_blocks[block_num*16],h);
    block_num=(1+block_num)%max_blocks;
    
    // if it is a header block, and it doesnt look
    // like all the data blocks are in the cache 
    // (suggesting it isnt just a local insert), then
    // add it to our list of 'interesting' blocks.
    if (block_type=='H')
    {
        interesting_block_time[interesting_block_num]=time(NULL);
        interesting_block_check_time[interesting_block_num]=0;
        write_key(interesting_blocks[interesting_block_num][0],h);
        write_key(interesting_blocks[interesting_block_num][1],last_block);
        interesting_block_num=(1+interesting_block_num)%MAX_INTERESTING_BLOCKS;
    }
    
    return 0;
}

// read a block from the block cache...
long read_block(unsigned long id[4],BLOCK *block,unsigned long *key)
{
    char filename[256],*hx=tohex(id);
    unsigned long err=0;
    
//printf("read_block ID=%s\n",tohex(id));

    sprintf(filename,"%s/%.2s/%s",cache_dir,hx,hx);
    FILE *fd=fopen(filename,"rb");
    if (!fd) return -1; // open failed!
    if (fread((void *)block,sizeof(*block),1,fd)!=1)
        err=1;
    fclose(fd);
    
    if (err)
        return -2; // read failed.
    
    // unscramble block after reading
    if (key)
        scramble_buffer(block->payload.data,sizeof(block->payload.data),key);
    copy_key(passkey,tmpkey);
    scramble_buffer((unsigned char *)block,
		            ((char *)&block->payload.data)-((char *)block),
					tmpkey);
    
    if (block->block_type!='H' && block->block_type!='D')
        return -999; // invalid block!
    return 0;
}

// convert a file into blocks and insert them into the cache,
// then advertise the file to the world.
long insert_file_into_cache(char *filename,void (*progress_cb)(char *text,long n,long max))
{
    BLOCK last_header,header,data;
    unsigned long prng_save[4],content_hash[4];
    
    long total=0;
    FILE *fd=fopen(filename,"rb");
    if (!fd)
        return -1; // couldnt open file for reading
    fseek(fd,0,SEEK_END);
    long filesize=ftell(fd);
	if (filesize<=0)
	{
		fclose(fd); // zero length files are not allowed.
		return -3;
	}
    fseek(fd,0,SEEK_SET);
	long total_blocks=(filesize+sizeof(data.payload.data)-1)/
		               sizeof(data.payload.data);
	long last_progress=0;
	unsigned char part_num=(unsigned char)0;

	while (total_blocks>0)
	{
	    long block_count=0;
		long data_blocks=total_blocks;
		if (data_blocks>sizeof(header.payload.block_ids)/16)
			data_blocks=sizeof(header.payload.block_ids)/16;

		// pass1: create the header block...
		memset(&header,0x00,sizeof(header));
		header.block_type='H';
		write_key(header.key,prng);
		copy_key(prng,prng_save);
		memset(&content_hash[0],0x00,sizeof(content_hash));

		long file_pos=ftell(fd);

		while (data_blocks-->0)
		{
			long bytes=fread(data.payload.data,1,sizeof(data.payload.data),fd);
			unsigned long h[4];
			memset(&data,0x00,sizeof(data)-sizeof(data.payload.data));
			data.block_type='D';
        
			if (bytes<sizeof(data.payload.data))
				memset(data.payload.data+bytes,0x00,sizeof(data.payload.data)-bytes);
        
			// crappy content hash mechanism with feedback... assumes
			// content_hash member of BLOCK immediately preceeds
			// payload member.
			write_key(data.content_hash,content_hash);
			hash(&data.content_hash[0],
				 sizeof(data.content_hash)+sizeof(data.payload),
				 content_hash);

			// just work out the id, dont actually write to disk
			// since we want to write the header block first.
			if (write_block(&data,prng,h,0)<0)
			{
				fclose(fd);
				return -2; // block write failed
			}
			write_key(header.payload.block_ids[block_count],h);
			block_count++;
			total+=bytes;
        
			if (progress_cb)
			{
				if (total-last_progress>(filesize+filesize)/200)
				{
					last_progress=total;
					progress_cb("Building header block...",total,filesize+filesize);
				}
			}
		}
    
		// description is filename without path
		char *fname=filename;
		while (strchr(fname,'/'))
			fname=strchr(fname,'/')+1;
		strcpy((char *)header.text,fname); 

		// store part number
		header.part[0]=part_num;

		// store content hash
		write_key(header.content_hash,content_hash);
    
		header.size[0]=(filesize>>24) & 0xff;
		header.size[1]=(filesize>>16) & 0xff;
		header.size[2]=(filesize>>8) & 0xff;
		header.size[3]=(filesize) & 0xff;
    
		unsigned long header_id[4];
		BLOCK header_copy;
		memcpy(&header_copy,&header,sizeof(header));
		write_block(&header_copy,NULL,header_id,0);
		if (part_num)
		{
			write_key(last_header.next_header,header_id);
			write_block(&last_header,NULL);
		}
		memcpy(&last_header,&header,sizeof(header));
    
		// pass2: then write the data blocks to disk...
		copy_key(prng_save,prng);    
		data_blocks=total_blocks;
		if (data_blocks>sizeof(header.payload.block_ids)/16)
			data_blocks=sizeof(header.payload.block_ids)/16;
		fseek(fd,file_pos,SEEK_SET);
		while (data_blocks-->0)
		{
			long bytes=fread(data.payload.data,1,sizeof(data.payload.data),fd);
			memset(&data,0x00,sizeof(data)-sizeof(data.payload.data));
			data.block_type='D';
        
			if (bytes<sizeof(data.payload.data))
				memset(data.payload.data+bytes,0x00,sizeof(data.payload.data)-bytes);
			total+=bytes;
        
			if (write_block(&data,prng)<0)
			{
				fclose(fd);
				return -6; // block write failed
			}
			if (progress_cb)
			{
				if (total-last_progress>(filesize+filesize)/200)
				{
					last_progress=total;
					progress_cb("Inserting blocks...",total,filesize+filesize);
				}
			}
			total_blocks--;
		}
		// advertise file to world...
		advertise_file(header_id,content_hash,fname,part_num,filesize);
		part_num++;

		// advance prng...
		char tmpbuf[100];
		strcpy(tmpbuf,"XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX");
		scramble_buffer((unsigned char *)tmpbuf,32,prng);
		hash((unsigned char *)tmpbuf,32,prng);
	}
    fclose(fd);
	write_block(&last_header,NULL);

    return 0;
}

// reconstruct a file from blocks in the block cache...
long create_file_from_blocks(unsigned long id[4],void (*progress_cb)(char *text,long n,long max))
{
    BLOCK header,data;
    
    if (read_block(id,&header)<0)
        return -1; // couldnt read block
    if (header.block_type!='H')
        return -2; // not a header block
    
    unsigned long key[4];
    long filesize=(header.size[0]<<24) +
        (header.size[1]<<16) +
        (header.size[2]<<8)  +
        (header.size[3]);
	long total_filesize=filesize,total_bytes=0;
    long block_count=0;
    long write_size;
    read_key(header.key,key);
    
    long total_data_blocks=(filesize+sizeof(header.payload.data)-1)/
                           sizeof(header.payload.data);
    
    char filename[256];
    sprintf(filename,"%s/../files",cache_dir);
    mkdir(filename,0777);
    sprintf(filename,"%s/../files/%s",cache_dir,(char *)header.text);
    FILE *fd=fopen(filename,"wb");
    if (!fd)
        return -3; // couldnt open file for writing
	long last_progress=0;
    while (filesize>0)
    {
        unsigned long h[4];
        long rc;
        read_key(header.payload.block_ids[block_count],h);
        if ((rc=read_block(h,&data,key))<0)
        {
            fclose(fd);
            return -4; // block read failed
        }
        
        write_size=sizeof(data.payload.data);
        if (write_size>filesize)
            write_size=filesize;
        
        fwrite(data.payload.data,write_size,1,fd);
        filesize-=write_size;
		total_bytes+=write_size;
        if (filesize && ++block_count>=sizeof(data.payload.block_ids)/16)
		{
			read_key(header.next_header,id);
			if (read_block(id,&header)<0)
			{
				fclose(fd);
				return -1; // couldnt read block
			}
			if (header.block_type!='H')
			{
				fclose(fd);
				return -2; // not a header block
			}
			read_key(header.key,key);
			block_count=0;
		}
			               
		if (progress_cb)
		{
			if (total_bytes-last_progress>total_filesize/200)
			{
				last_progress=total_bytes;
				progress_cb("Writing file...",total_bytes,total_filesize);
			}
		}
    }
    fclose(fd);

	// advance prng...
	char tmpbuf[100];
	strcpy(tmpbuf,"XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX");
    scramble_buffer((unsigned char *)tmpbuf,32,prng);
	hash((unsigned char *)tmpbuf,32,prng);
        
    return 0;
}

// remove all directories and all files below path
// same as rm -r <path> but does some basic filename
// vadlidation (after I lost my hard disk last time :-/ )
static void zapdir(char *path)
{
    if (!path)
        return;
    if (!*path)
        return;
    long q;
#ifdef WIN32
    char dirpath[1024];
    sprintf(dirpath,"%s/*",path);
    WIN32_FIND_DATA fdata;
    HANDLE fh=FindFirstFile(dirpath,&fdata);
    if (fh==INVALID_HANDLE_VALUE)
        return; // FindFirstFile failed
    do
    {
        if (strcmp(fdata.cFileName,".")==0 ||
            strcmp(fdata.cFileName,"..")==0)
            continue;
        char fpath[1024];
        sprintf(fpath,"%s/%s",path,fdata.cFileName);
        if (fdata.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
        {
            // make sure we only delete cache directories...
            if (strlen(fdata.cFileName)!=2)
                return; // cache directories are 2 hex digits
            if (!isxdigit(fdata.cFileName[0]) || !isxdigit(fdata.cFileName[1]))
                return; // cache directories are 2 hex digits
            zapdir(fpath);
            continue;
        }
        
        // make sure we only delete cache files...
        if (strlen(fdata.cFileName)!=32)
            return; // eh? file should be 32 hex digits?
        for (q=0;q<32;q++)
        {
            if (!isxdigit(fdata.cFileName[q]))
                return; // ERROR: file should be 32 hex digits??
        }
        
        if (unlink(fpath)<0)
            return; // couldnt delete file
    } while (FindNextFile(fh,&fdata));
    FindClose(fh);
    rmdir(path);
    
#else
    DIR *dirp=opendir(path);
    struct dirent *dp;
    if (!dirp) return; // invalid path
    while ((dp=readdir(dirp))!=NULL)
    {
        if (strcmp(dp->d_name,".")==0 ||
            strcmp(dp->d_name,"..")==0)
            continue;
        
        char filename[2048];
        sprintf(filename,"%s/%s",path,dp->d_name);
        struct stat buf;
        if (stat(filename,&buf)<0) return;    // cant stat()
        if (S_ISDIR(buf.st_mode))
        {
            // make sure we only delete cache directories...
            if (strlen(dp->d_name)!=2)
                return; // cache directories are 2 hex digits
            if (!isxdigit(dp->d_name[0]) || !isxdigit(dp->d_name[1]))
                return; // cache directories are 2 hex digits
            zapdir(filename);
            continue;
        }
        // make sure we only delete cache files...
        if (strlen(dp->d_name)!=32)
            return; // eh? file should be 32 hex digits?
        for (q=0;q<32;q++)
        {
            if (!isxdigit(dp->d_name[q]))
                return; // ERROR: file should be 32 hex digits??
        }
        if (unlink(filename)<0)
            return; // couldnt delete file
    }
    closedir(dirp);
    rmdir(path);
#endif
}

// create the cache directory structure
long create_initial_cache(long maximum_blocks)
{
    // destroy any previous cache directory structure...
    destroy_cache();
    
    // try to build the cache directory structure
    printf("creating cache directory structure...\n");
    mkdir(cache_dir,0777);
    erase_timestamp(cache_dir);
    long q;
    for (q=0;q<256;q++)
    {
        char dirname[1024];
        sprintf(dirname,"%s/%02lX",cache_dir,q);
        mkdir(dirname,0777);
        erase_timestamp(dirname);
    }
    if (access(cache_dir,R_OK|W_OK)!=0)
    {
        printf("   ... could not create\n");
        return -1; // cache dir not created or premissions problems
    }
    
    // initialise block list
    if (!(all_blocks=(unsigned char *)calloc(1,maximum_blocks*16)))
        return -2; // not enough memory
    block_num=0;
    max_blocks=maximum_blocks;
    memset(interesting_block_time,0x00,sizeof(interesting_block_time));
    memset(interesting_blocks,0x00,sizeof(interesting_blocks));
    interesting_block_num=0;
    
    return 0;
}

void destroy_cache()
{
    printf("deleteing any existing block cache structure...\n");
    zapdir(cache_dir);

    // free up memory for block list
    if (all_blocks)
    {
        free(all_blocks);
        all_blocks=NULL;
    }
}

void check_interesting_blocks()
{
    long num_reads=8; // absolute max number of ads this run....
    time_t t=time(NULL);
    long endpos=interesting_block_num;
    long pos=interesting_block_num;
    
    do
    {
        if (interesting_block_time[pos])
        {
            if (t-interesting_block_time[pos]>600)
            {
                // give up on blocks older than 10mins
                interesting_block_time[pos]=0;
            }
            else if (t-interesting_block_check_time[pos]>=120)
            {
                // every 2 mins check to see if we have the
                // whole file...
                interesting_block_check_time[pos]=t;
                
                // check to see if the last data block is in the cache...
                unsigned long header_id[4],id[4],content_hash[4];
                read_key(interesting_blocks[pos][1],id);
                char filename[256],*hx=tohex(id);
                sprintf(filename,"%s/%.2s/%s",cache_dir,hx,hx);
                FILE *fd=fopen(filename,"rb");
                if (fd)
                {
                    // ok, found last data block, it should be
                    // alright to advertise the file.
                    fclose(fd);
                    
                    read_key(interesting_blocks[pos][0],header_id);
                    BLOCK header;
                    if (read_block(header_id,&header)==0)
                    {
                        long filesize=(header.size[0]<<24) +
                            (header.size[1]<<16) +
                            (header.size[2]<<8)  +
                            (header.size[3]);
						read_key(header.content_hash,content_hash);
                        advertise_file(header_id,content_hash,
							           (char *)header.text,
									   header.part[0],filesize);
                    }
                    num_reads--;
                    
                    // clear entry in 'interesting blocks' list.
                    interesting_block_check_time[pos]=0;
                }
            }
        }
        pos=(1+pos)%MAX_INTERESTING_BLOCKS;
    } while (pos!=endpos && num_reads>0);
}
