
#include "sock.h"


/*** initialise sockets interface (unlimit file descriptors) ***/
static long net_init_lev=0;
long NET_initiate()
{
    if (net_init_lev==0)
    {
#ifdef WIN32
        unsigned short wVersionRequested = MAKEWORD(1, 1);
        WSADATA wsadata;
        int err;
        /* Need to include library: wsock32.lib for Windows Sockets */
        err = WSAStartup(wVersionRequested, &wsadata);
        if (err) return -1; // hmmm... we're stuffed!
#else
#ifdef RLIMIT_NOFILE
        struct rlimit rlp;
        if (!getrlimit(RLIMIT_NOFILE,&rlp))
        {
            rlp.rlim_cur=rlp.rlim_max>=FD_SETSIZE?FD_SETSIZE-1:rlp.rlim_max;
            setrlimit(RLIMIT_NOFILE,&rlp);
        }
#endif
#endif
    }
#ifdef SIGPIPE
    sigignore(SIGPIPE);
#endif
    net_init_lev++;
    return 0;
}

/*** terminate sockets interface ***/
long NET_terminate()
{
    if ((--net_init_lev)==0)
    {
#ifdef WIN32
        WSACleanup();
#endif
    }
    return 0;
}

/*** convert service name/number string to port number ***/
long NET_service_to_port(const char *service)
{
    struct servent *sp=getservbyname(service,"tcp");
    if (sp==NULL)
    {
        if (isdigit(service[0]))
            return atoi(service);
        return -1;
    }
    return sp->s_port;
}

/*** convert hostname to aaa.bbb.ccc.ddd ip address ***/
char *NET_host_to_ip(const char *hostname)
{
    struct hostent *h;
    struct in_addr in;
    static char ip[256];
    if (!hostname) { strcpy(ip,"localhost"); hostname=ip; }
    if (!hostname[0]) { strcpy(ip,"localhost"); hostname=ip; }
    if (inet_addr(hostname)!=INADDR_NONE)
    {
        strcpy(ip,hostname);
        return ip;
    }
    if (!(h=gethostbyname(hostname)))
        return NULL;
    memcpy(&in.s_addr,*h->h_addr_list,sizeof(in.s_addr));
    strcpy(ip,(char *)inet_ntoa(in));
    return ip;
}

/*** convert aaa.bbb.ccc.ddd ip address to hostname ***/
char *NET_ip_to_host(const char *ip)
{
    struct hostent *h;
    long addr;
    static char hostname[256];
    if ((addr=inet_addr(ip))==INADDR_NONE)
        return NULL;
    if (!(h=gethostbyaddr((char *)&addr,sizeof(addr),AF_INET)))
        return NULL;
    strcpy(hostname,h->h_name);
    return hostname;
}

/*** make a socket non-blocking ***/
long NET_make_nonblocking(const long fd)
{
#ifdef WIN32
    unsigned long arg = 1;
    int opt=1;
    setsockopt(fd,SOL_SOCKET,SO_KEEPALIVE,(char *)&opt,sizeof(opt));
    return ::ioctlsocket(fd, FIONBIO, &arg);
#else
    int opt=1;
    setsockopt(fd,SOL_SOCKET,SO_KEEPALIVE,(char *)&opt,sizeof(opt));
#ifdef O_NONBLOCK
    return fcntl(fd,F_SETFL,O_NONBLOCK);
#else
    return fcntl(fd,F_SETFL,O_NDELAY);
#endif
#endif
}

/*** close a socket ***/
long NET_close(const long fd)
{
#ifdef WIN32
    long err=errno;
    return ::closesocket(fd);
#else
    return close(fd);
#endif
}

/*** connect a socket ***/
long NET_connect(const char *ip,const long port,long block_on_connect)
{
    long fd;
    struct sockaddr_in serv_addr;
    memset((char *)&serv_addr,0x00,sizeof(serv_addr));
    serv_addr.sin_family      = AF_INET;
    serv_addr.sin_addr.s_addr = inet_addr(ip);
    serv_addr.sin_port        = htons((unsigned short)port);
    if ((fd=socket(AF_INET,SOCK_STREAM,0))<0)
        return fd;
    if (!block_on_connect)
    {
        if (NET_make_nonblocking(fd)<0)
        { NET_close(fd); return fd=-101; }
    }
    if (connect(fd,(struct sockaddr *)&serv_addr,sizeof(serv_addr))<0)
    {
        if (NET_bad_errno())
        {
            NET_close(fd); return fd=-100;
        }
    }
    if (NET_make_nonblocking(fd)<0)
    { NET_close(fd); return fd=-101; }
    return fd;
}

/*** passive listen on a port ***/
long NET_listen(const long port)
{
    int opt=1;
    long fd;
    struct sockaddr_in server;
    memset(&server,0x00,sizeof(server));
    if ((fd=socket(AF_INET,SOCK_STREAM,0))<0) return fd;
    if (setsockopt(fd,SOL_SOCKET,SO_REUSEADDR,(char *)&opt,sizeof(opt))<0)
    { NET_close(fd); return fd=-100; }
    server.sin_family=AF_INET;
    server.sin_addr.s_addr=htonl(INADDR_ANY);
    server.sin_port=htons((unsigned short)port);
    if (bind(fd,(struct sockaddr *)&server,sizeof(server)))
    { NET_close(fd); return fd=-101; }
    if (NET_make_nonblocking(fd)==-1)
    { NET_close(fd); return fd=-102; }
    listen(fd,5);
    return fd;
}

/*** accept a connection on a port ***/
long NET_accept(const long fd,char *fromip /*** 16 char buffer ***/ )
{
    struct sockaddr_in addr;
    int addr_len=sizeof(addr);
    long s;
    fromip[0]=0;
    s=accept(fd,(struct sockaddr *) &addr, &addr_len);
    if (s>=0)
    {
        strcpy(fromip,(char *)inet_ntoa(addr.sin_addr));
        if (NET_make_nonblocking(s)==-1)
        { NET_close(s); return -102; }
    }
    return s;
}

/*** read some data from a socket ***/
long NET_read(const long fd,const char *buf,const long bufsize)
{
#ifdef WIN32
    return ::recv(fd,(char *)buf,bufsize,0);
#else
    return read(fd,(char *)buf,bufsize);
#endif
}

/*** write some data to a socket ***/
long NET_write(const long fd,const char *buf,const long bufsize)
{
#ifdef WIN32
    return ::send(fd,buf,bufsize,0);
#else
    return write(fd,buf,bufsize);
#endif
}

int NET_bad_errno()
{
    long err=errno;
    if (
#ifdef EWOULDBLOCK
        err==EWOULDBLOCK ||
#endif
#ifdef EPIPE
        err==EPIPE ||
#endif
#ifdef EINTR
        err==EINTR || 
#endif
#ifdef EAGAIN
        err==EAGAIN || 
#endif
#ifdef EINPROGRESS
        err==EINPROGRESS ||
#endif
#ifdef ENOTCONN
        err==ENOTCONN ||
#endif
        0)
        return 0;
    return 1;
}

//-----------------------------------------------------
//-----------------------------------------------------
//-----------------------------------------------------

NET_SOCKET NET_sockets[NET_MAX_SOCKETS];

static long creation_id=1; // just a counter which increments

static long NET_current_sockets=0;
static long max_sockets=20;
static long max_descriptor=0;
static fd_set input_set,output_set,except_set;
static void (*NET_accept_connection_callback)(long socket,void *user_data)=NULL;
static void (*NET_process_input_callback)(long socket,char *ip,void *user_data,
                                          char *buf, long bufsize)=NULL;
static void (*NET_process_exception_callback)(long socket,void *user_data)=NULL;

long NET_fileno(long s)
{
    if (s<0 || s>=NET_MAX_SOCKETS) return -1;
    return NET_sockets[s].fd;
}

long NET_findsock(long fd)
{
    long n=NET_current_sockets;
    long s;
    for (s=0;n>0 && s<NET_MAX_SOCKETS;s++)
    {
        if (NET_sockets[s].type)
        {
            if (NET_sockets[s].fd==fd)
                return s;
            n--;
        }
    }
    return -1; /* not found */
}

static long NET_getsock(long fd)
{
    long q;
    if (fd<0) return fd;
    for (q=0;q<NET_MAX_SOCKETS;q++)
    {
        if (!NET_sockets[q].type)
        {
            NET_current_sockets++;
            NET_sockets[q].fd=fd;
            NET_sockets[q].NET_accept_connection_callback=NET_accept_connection_callback;
            NET_sockets[q].NET_process_input_callback=NET_process_input_callback;
            NET_sockets[q].NET_process_exception_callback=NET_process_exception_callback;
            return q;
        }
    }
    return -999;
}

/************************/
/*** destroy a socket ***/
/************************/
void NET_destroy_socket(long s)
{
    if (NET_sockets[s].type)
    {
        NET_close(NET_fileno(s));
        if (NET_sockets[s].send_buffer!=NULL)
            free(NET_sockets[s].send_buffer);
        NET_sockets[s].type=0;
        NET_current_sockets--;
    }
    memset(NET_sockets+s,0x00,sizeof(NET_sockets[s]));
}

/*********************************************/
/*** Initialise socket array.              ***/
/*** Set the max number of sockets         ***/
/*** which this process can user.          ***/
/*** <0 => Initialisation failed!          ***/
/*********************************************/
static long net_init_count=0;
long NET_init(void (*accept)(long,void *),
              void (*input)(long,char *,void *,char *,long),
              void (*except)(long,void *))
{
    if (++net_init_count>1)
        return max_sockets;
    max_sockets=NET_initiate();
    memset(NET_sockets,0x00,sizeof(NET_sockets));
    NET_accept_connection_callback=accept;
    NET_process_input_callback=input;
    NET_process_exception_callback=except;
    return max_sockets;
}

/*******************************/
/*** destroy all the sockets ***/
/*******************************/
void NET_term()
{
    if (--net_init_count)
        return;
    long s;
    for (s=0;s<NET_MAX_SOCKETS;s++)
    {
        if (NET_sockets[s].type)
            if (NET_sockets[s].NET_process_exception_callback)
                NET_sockets[s].NET_process_exception_callback(s,NET_sockets[s].user_data);
            NET_destroy_socket(s);
    }
    max_sockets=20;
    NET_current_sockets=0;
    max_descriptor=0;
}


/*******************************************/
/*** create a listening socket on a port ***/
/*******************************************/
long NET_create_listening_socket(long port,void *user_data)
{
    long opt=1;
    long fd=NET_listen(port);
    if (fd<0)
        return fd;
    long s=NET_getsock(fd);
    if (s>=0)
    {
        NET_sockets[s].type=1; 
        NET_sockets[s].id=creation_id++;
        NET_sockets[s].port=port;
        NET_sockets[s].user_data=user_data;
        NET_sockets[s].connected=1;
    }
    return s;
}

/***********************************************/
/*** accept connection to a listening socket ***/
/***********************************************/
static long accept_connection(long listening_sock)
{
    char fromip[20];
    long s,fd;
    char *ident="";
    if ((fd=NET_accept(NET_fileno(listening_sock),fromip))<0)
        return fd;
    s=NET_getsock(fd);
    if (s>=0)
    {
        NET_sockets[s].type=2;
        NET_sockets[s].id=creation_id++;
        NET_sockets[s].port=NET_sockets[listening_sock].port;
        strcpy(NET_sockets[s].ip,fromip);
        NET_sockets[s].send_buffer_size=0;
    }
    return s;
}

/**********************************/
/*** connect to a remote server ***/
/**********************************/
long NET_connect_remote(char *ip, long port, void *user_data)
{
    long s,fd;
    if ((fd=NET_connect(ip,port,0))<0)
        return fd;
    if ((s=NET_getsock(fd))<0)
    {
        NET_close(fd);
        return -999;
    }
    NET_sockets[s].type=2;
    NET_sockets[s].id=creation_id++;
    strcpy(NET_sockets[s].ip,ip);
    NET_sockets[s].port=port;
    NET_sockets[s].user_data=user_data;
    NET_sockets[s].connected=0;
    return s;
}

/************************/
/*** populate fd_sets ***/
/************************/
static void make_fd_sets()
{
    long s,p;
    FD_ZERO(&input_set);
    FD_ZERO(&output_set);
    FD_ZERO(&except_set);
    max_descriptor=0;
    for (s=p=0;p<NET_current_sockets && s<NET_MAX_SOCKETS;s++)
    {
        if (NET_sockets[s].type)
        {
            if (NET_sockets[s].connected)
                FD_SET(NET_fileno(s),&input_set);
            if (NET_sockets[s].send_buffer_size!=0 ||
                !NET_sockets[s].connected)
            {
                FD_SET(NET_fileno(s),&output_set);
            }
            FD_SET(NET_fileno(s),&except_set);
            p++;
            if (NET_fileno(s)>max_descriptor)
                max_descriptor=NET_fileno(s);
        }
    }
    
}


/********************************/
/*** try to flush queued data ***/
/********************************/
void NET_send_now(long s)
{
    long cnt,sz;
    if (NET_sockets[s].send_buffer_size==0 || NET_sockets[s].connected==0)
        return;
    sz=NET_sockets[s].send_buffer_size;
    if (sz>512)
        sz=512; // write in little blocks
    cnt=NET_write(NET_fileno(s),NET_sockets[s].send_buffer,sz);
    if (cnt<0)
    {
        if (NET_bad_errno())
        {
            if (NET_sockets[s].NET_process_exception_callback)
                NET_sockets[s].NET_process_exception_callback(s,NET_sockets[s].user_data);
            NET_destroy_socket(s);
        }
        NET_sockets[s].last_block_time=(long)time(NULL);
        return;
    }
    NET_sockets[s].last_block_time=0;
    if (NET_sockets[s].send_buffer_size-=cnt)
        NET_sockets[s].last_block_time=(long)time(NULL);
    
    if (NET_sockets[s].send_buffer_size==0)
    {
        free(NET_sockets[s].send_buffer);
        NET_sockets[s].send_buffer=NULL;
    }
    else if (NET_sockets[s].send_buffer_size>0)
    {
        memcpy(NET_sockets[s].send_buffer,
            NET_sockets[s].send_buffer+cnt,
            NET_sockets[s].send_buffer_size);
    }
}

static long last_sock_checked=NET_MAX_SOCKETS;
static long throttle_time=0,throttle_bandwidth=16777216,
throttle_read=0,throttle_write=0;
/*******************************************/
/* set global read/write throttle limits ***/
/*******************************************/
void NET_set_throttle(long bytes_per_sec)
{
    if (bytes_per_sec>0)
    {
        throttle_time=0;
        throttle_bandwidth=bytes_per_sec;
    }
}
long NET_get_throttle()
{
    return throttle_bandwidth;
}


/**********************************************/
/*** check sockets for input and exceptions ***/
/**********************************************/
long NET_do_select(unsigned long timeout_u_sec)
{
    long found;
    struct timeval timeout;
    make_fd_sets();
    memset(&timeout,0x00,sizeof(timeout));
    timeout.tv_sec=timeout_u_sec/1000000;
    timeout.tv_usec=timeout_u_sec%1000000;
    if ((found=select(max_descriptor+1,&input_set,&output_set,&except_set,
        &timeout))<0)
    {
        if (NET_bad_errno()) return -1;
    }
    else if (found)
    {
        long n,s=last_sock_checked,t=(long)time(NULL);
        if (t!=throttle_time)
        {
            throttle_time=t;
            throttle_read=throttle_bandwidth;
            throttle_write=throttle_bandwidth;
        }
        if (throttle_read<=0 || throttle_write<=0)
        {
            // sleep for 1/100th of a sec
            // so we're CPU friendly when throttling.
            NET_sleep(10000); 
        }
        for (n=0;n<NET_MAX_SOCKETS && found; n++)
        {
            if (++s>=NET_MAX_SOCKETS)
                s=0;
            
            if (FD_ISSET(NET_fileno(s),&except_set))
            {
                if (NET_sockets[s].NET_process_exception_callback)
                    NET_sockets[s].NET_process_exception_callback(s,NET_sockets[s].user_data);
                NET_destroy_socket(s);
                found--;
                continue;
            }
            if (FD_ISSET(NET_fileno(s),&input_set))
            {
                if (NET_sockets[s].type==1)
                {
                    long ns=accept_connection(s);
                    if (ns>=0 && NET_sockets[s].NET_accept_connection_callback)
                        NET_sockets[s].NET_accept_connection_callback(ns,NET_sockets[s].user_data);
                    found--;
                }
                else if (throttle_read>0 && NET_sockets[s].type==2)
                {
                    long n,sz=512;
                    char buf[512]; // read in little blocks
                    if (sz>throttle_read)
                        sz=throttle_read;
                    n=NET_read(NET_fileno(s),buf,sz);
                    if (n>0)
                    {
                        if (NET_sockets[s].NET_process_input_callback)
                            NET_sockets[s].NET_process_input_callback(s,NET_sockets[s].ip,
                            NET_sockets[s].user_data,
                            buf,n);
                        throttle_read-=n;
                    }
                    else if (n==0 || NET_bad_errno())
                    {
                        if (NET_sockets[s].NET_process_exception_callback)
                            NET_sockets[s].NET_process_exception_callback(s,NET_sockets[s].user_data);
                        NET_destroy_socket(s);
                    }
                    found--;
                }
            }
            if (FD_ISSET(NET_fileno(s),&output_set))
            {
                NET_sockets[s].connected=1;
                if (throttle_write>0 && NET_sockets[s].send_buffer_size)
                {
                    long sz=NET_sockets[s].send_buffer_size;
                    NET_send_now(s);
                    
                    throttle_write-=(sz-NET_sockets[s].send_buffer_size);
                    if (NET_sockets[s].last_block_time)
                    {
                        if (--s<0) s=NET_MAX_SOCKETS;
                        break; // connection is blocked, start here next time
                    }
                }
                found--;
            }
            
            // break connections which have been write blocked for more than 90secs
            if (NET_sockets[s].type && NET_sockets[s].send_buffer_size>0 &&
                NET_sockets[s].last_block_time)
            {
                if (t-NET_sockets[s].last_block_time>90)
                {
                    if (NET_sockets[s].NET_process_exception_callback)
                        NET_sockets[s].NET_process_exception_callback(s,NET_sockets[s].user_data);
                    NET_destroy_socket(s);
                }
            }
        }
        last_sock_checked=s;
    }
    return 0;
}

/**********************************************/
/*** queue up data to be sent down a socket ***/
/**********************************************/
void NET_send(long s,char *data, long size)
{
    if (NET_sockets[s].send_buffer==NULL)
        NET_sockets[s].send_buffer=(char *)malloc(size);
    else
    {
        NET_sockets[s].send_buffer=(char *)realloc(NET_sockets[s].send_buffer,
            NET_sockets[s].send_buffer_size+size);
    }
    memcpy(NET_sockets[s].send_buffer+NET_sockets[s].send_buffer_size,data,size);
    NET_sockets[s].send_buffer_size+=size;
}

/********************************************/
/*** portable sleep function              ***/
/*** NB: timeout is in millionths of secs ***/
/********************************************/
void NET_sleep(long u_secs)
{
#ifdef WIN32
    Sleep(1+(u_secs/1000));
#else
    usleep(u_secs);
#endif
    /***
    struct timeval timeout;
    memset(&timeout,0x00,sizeof(timeout));
    timeout.tv_sec=u_secs/1000000;
    timeout.tv_usec=u_secs%1000000;
    select(1,NULL,NULL,NULL,&timeout);
    ***/
}

